@@ -27,9 +27,11 @@ int remember_for_completion = FALSE;    /* whether we should put al words from i
 char *multiline_separator = NULL;       /* character sequence to use in lieu of newline when storing multi-line input in a single history line */
 char *pre_given = NULL;                 /* pre-given user input when rlwrap starts up */
 struct rl_state saved_rl_state = { "", "", 0, 0, 0 };      /* saved state of readline */
+bool bracketed_paste_enabled = FALSE;
 static char return_key;                 /* Key pressed to enter line */
 static int forget_line;
 static char *colour_start, *colour_end;        /* colour codes */
+
 int multiline_prompts = TRUE;
 
 /* forward declarations */
@@ -39,12 +41,13 @@ static int my_accept_line(int, int);
 static int my_accept_line_and_forget(int, int);
 static void munge_file_in_editor(const char *filename, int lineno, int colno);
 static Keymap getmap(const char *name);
-static void bindkey(int key, rl_command_func_t *function, const char *maplist);
-
+static void bindkey(int key, rl_command_func_t *function, const char *function_name, const char *maplist);
+#define RL_COMMAND_FUN(f) f, #f
 
 /* readline bindable functions: */
 static int munge_line_in_editor(int, int);
 static int direct_keypress(int, int);
+static int direct_prefix(int, int);
 static int handle_hotkey(int, int);
 static int handle_hotkey_without_history(int, int);
 static int please_update_alaf(int,int);
@@ -54,6 +57,8 @@ static int please_update_ce(int,int);
 static int debug_ad_hoc(int,int);
 static int dump_all_keybindings(int,int);
 
+
+
 void
 init_readline(char *UNUSED(prompt))
 {
@@ -62,25 +67,31 @@ init_readline(char *UNUSED(prompt))
   rl_add_defun("rlwrap-accept-line-and-forget", my_accept_line_and_forget,-1);
   rl_add_defun("rlwrap-call-editor", munge_line_in_editor, -1);
   rl_add_defun("rlwrap-direct-keypress", direct_keypress, -1);  
+  rl_add_defun("rlwrap-direct-prefix", direct_prefix, -1);
   rl_add_defun("rlwrap-hotkey", handle_hotkey, -1);
   rl_add_defun("rlwrap-hotkey-without-history", handle_hotkey_without_history, -1);
 
   /* only useful while debugging */
   rl_add_defun("rlwrap-dump-all-keybindings", dump_all_keybindings,-1);
   rl_add_defun("rlwrap-debug-ad-hoc", debug_ad_hoc, -1);
+
+  /* if someone's .inputrc binds a key to accept-line, make it use our own version in lieu of readline's */
+  rl_add_defun("accept-line", my_accept_line, -1);
+  
+
   
   /* the old rlwrap bindable function names with underscores are deprecated: */
   rl_add_defun("rlwrap_accept_line_and_forget", please_update_alaf,-1);
   rl_add_defun("rlwrap_call_editor", please_update_ce,-1);
   
+  /* put the next variable binding(s) *before* rl_initialize(), so they can be overridden */
+  rl_variable_bind("blink-matching-paren","on");  
   
-  rl_variable_bind("blink-matching-paren","on"); /* Shouldn't this be on by default? */
-
-  bindkey('\n', my_accept_line, "emacs-standard; vi-insert; vi-command"); 
-  bindkey('\r', my_accept_line, "emacs-standard; vi-insert; vi-command"); 
-  bindkey(15, my_accept_line_and_forget, "emacs-standard; vi-insert; vi-command");	/* ascii #15 (Control-O) is unused in readline's emacs and vi keymaps */
+  bindkey('\n', RL_COMMAND_FUN(my_accept_line), "emacs-standard; vi-insert; vi-command"); 
+  bindkey('\r', RL_COMMAND_FUN(my_accept_line), "emacs-standard; vi-insert; vi-command"); 
+  bindkey(15, RL_COMMAND_FUN(my_accept_line_and_forget), "emacs-standard; vi-insert; vi-command");	/* ascii #15 (Control-O) is unused in readline's emacs and vi keymaps */
   if (multiline_separator) 
-    bindkey(30, munge_line_in_editor, "emacs-standard;vi-insert;vi-command");            /* CTRL-^: unused in vi-insert-mode, hardly used in emacs  (doubles arrow-up) */
+    bindkey(30, RL_COMMAND_FUN(munge_line_in_editor), "emacs-standard;vi-insert;vi-command");           /* CTRL-^: unused in vi-insert-mode, hardly used in emacs  (doubles arrow-up) */
 
   
   
@@ -90,7 +101,11 @@ init_readline(char *UNUSED(prompt))
 				   readline will not bother to call tgetent(), will be agnostic about terminal
 				   capabilities and hence not be able to honour e.g. a set horizontal-scroll-mode off
 				   in .inputrc */
-  
+
+  /* put the next variable binding(s) *after* rl_initialize(), so they cannot be overridden */
+  /* rl_variable_bind("enable-bracketed-paste","off"); */     /* enable-bracketed-paste changes cursor positioning after printing the prompt ...
+                                                           ... causing rlwrap to overwrite it after accepting input                        */
+ 
   using_history();
   rl_redisplay_function = my_redisplay;
   rl_completion_entry_function =
@@ -98,6 +113,7 @@ init_readline(char *UNUSED(prompt))
   
   rl_catch_signals = FALSE;
   rl_catch_sigwinch = FALSE;
+
   saved_rl_state.input_buffer = mysavestring(pre_given ? pre_given : ""); /* Even though the pre-given input won't be displayed before the first 
                                                                              cooking takes place, we still want it to be accepted when the user 
                                                                              presses ENTER before that (e.g. because she already knows the 
@@ -105,14 +121,20 @@ init_readline(char *UNUSED(prompt))
   saved_rl_state.point = strlen(saved_rl_state.input_buffer);
   saved_rl_state.raw_prompt = mysavestring("");
   saved_rl_state.cooked_prompt = NULL;
-  
+
+  bracketed_paste_enabled = term_enable_bracketed_paste != NULL && strings_are_equal(rl_variable_value("enable-bracketed-paste"),"on");
+  DPRINTF2(DEBUG_READLINE, "var: %s, enabled: %s", rl_variable_value("enable-bracketed-paste"), term_enable_bracketed_paste);
+  if (bracketed_paste_enabled) {
+    DPRINTF0(DEBUG_READLINE, "bracketed-paste is enabled");
+    my_putstr(term_enable_bracketed_paste);
+  }
 }
 
 
 /* save readline internal state in rl_state, redisplay the prompt
    (so that client output gets printed at the right place) */
 void
-save_rl_state()
+save_rl_state(void)
 {
   free(saved_rl_state.input_buffer); /* free(saved_rl_state.raw_prompt) */;
   saved_rl_state.input_buffer = mysavestring(rl_line_buffer);
@@ -128,27 +150,26 @@ save_rl_state()
   my_redisplay();               /* and redisplay (this time without user input, cf the comments for the line_handler() function below) */
   rl_callback_handler_remove(); /* restore original terminal settings */
   rl_deprep_terminal();
-
 }
 
 
 /* Restore readline internal state from rl_state.   */
 
 void
-restore_rl_state()
+restore_rl_state(void)
 {
   
   char *newprompt;
-
-
   move_cursor_to_start_of_prompt(impatient_prompt ? ERASE : DONT_ERASE);
 
   cook_prompt_if_necessary();
-  newprompt =  mark_invisible(saved_rl_state.cooked_prompt); /* bracket (colour) control sequences with \001 and \002 */  
+  newprompt =  mark_invisible(saved_rl_state.cooked_prompt); /* bracket (colour) control sequences with \001 and \002 */
+  if (!strings_are_equal(newprompt, saved_rl_state.cooked_prompt))
+    DPRINTF2(DEBUG_READLINE, "newprompt: <%s>, cooked prompt: <%s>", M(newprompt), M(saved_rl_state.cooked_prompt));
   rl_expand_prompt(newprompt);
   mirror_slaves_echo_mode();    /* don't show passwords etc */
   
-  DPRINTF1(DEBUG_READLINE,"newprompt now %s", mangle_string_for_debug_log(newprompt,MANGLE_LENGTH));
+  DPRINTF1(DEBUG_READLINE,"newprompt now %s", M(newprompt));
   rl_callback_handler_install(newprompt, &line_handler);
   DPRINTF0(DEBUG_AD_HOC, "freeing newprompt");
   free(newprompt);             /* readline docs don't say it, but we can free newprompt now (readline apparently
@@ -167,7 +188,7 @@ void
 message_in_echo_area(char *message)
 {
   static int message_in_echo_area = FALSE;
-  DPRINTF1(DEBUG_READLINE, "message: %s", mangle_string_for_debug_log(message, MANGLE_LENGTH));
+  DPRINTF1(DEBUG_READLINE, "message: %s", M(message));
   if (message) {
     rl_save_prompt();
     message_in_echo_area = TRUE;  
@@ -180,6 +201,11 @@ message_in_echo_area(char *message)
   }     
 } 
 
+#ifndef HAVE_RL_VARIABLE_VALUE /* very ancient readline? */
+#  define rl_variable_value(s) "off"
+#endif
+
+
 static void
 line_handler(char *line)
 {
@@ -190,11 +216,12 @@ line_handler(char *line)
     /* colour_the_prompt = FALSE; don't mess with the cruft that may come out of dying command @@@ but command may not die!*/
     write_EOF_to_master_pty();
   } else {
+    /* NB: with bracketed-paste, "line" may actually contain newline characters */
     if (*line &&                 /* forget empty lines  */
         redisplay &&             /* forget passwords    */
         !forget_line &&          /* forget lines entered by CTRL-O */
         !match_regexp(line, forget_regexp, TRUE)) {     /* forget lines (case-inseitively) matching -g option regexp */ 
-      my_add_history(line);
+      my_add_history(line); /* if line consists of multiple lines, each line is added to history separately. Is this documented somewhere? */
     }
     forget_line = FALSE; /* until CTRL-O is used again */
 
@@ -209,13 +236,21 @@ line_handler(char *line)
        I think this method is more natural for multi-line input as well, (we will actually see our multi-line input as
        multiple lines) but not everyone will agree with that.
           
-       O.K, we know for sure that cursor is at start of line. When clients output arrives, it will be printed at
-       just the right place - but first we 'll erase the user input (as it may be about to be changed by the filter) */
-    
+       O.K, we know for sure that cursor is at start of line (after the prompt, or at position 0, if bracketed paste is
+       enabled)). When clients output arrives, it will be printed at just the right place - but first we 'll erase the
+       user input (as it may be about to be changed by the filter) */
+
+    if (bracketed_paste_enabled) {   /* we're at start of line and the prompt has just been erased */
+      SHOWCURSOR('1');
+      move_cursor_to_start_of_prompt(FALSE);
+      SHOWCURSOR('2');
+      my_putstr(saved_rl_state.cooked_prompt);
+      SHOWCURSOR('3');
+    }
+      
     rl_delete_text(0, rl_end);  /* clear line  (after prompt) */
     rl_point = 0;
     my_redisplay();             /* and redisplay (this time without user input, cf the comments for the line_handler() function below) */
-
     rewritten_line =
       (multiline_separator ? 
        search_and_replace(multiline_separator, "\n", line, 0, NULL,
@@ -245,7 +280,7 @@ line_handler(char *line)
     /*OK, now feed line to underlying command and wait for the echo to come back */
     put_in_output_queue(filtered_line);
     DPRINTF2(DEBUG_READLINE, "putting %d bytes %s in output queue", (int) strlen(rewritten_line),
-             mangle_string_for_debug_log(rewritten_line, 40));
+             M(rewritten_line));
     write_EOL_to_master_pty(return_key ? &return_key : "\n");
 
     accepted_lines++;
@@ -272,6 +307,11 @@ line_handler(char *line)
 
     if (one_shot_rlwrap)
       write_EOF_to_master_pty();
+
+    /* readline only outputs term_enable_bracketed_paste when we call rl_prep_terminal(). That's too   */
+    /* late for us, as we only call rl_prep_terminal *after* we have received user input               */
+    if (bracketed_paste_enabled)
+      my_putstr(term_enable_bracketed_paste);
   }
 }
 
@@ -311,39 +351,47 @@ static void
 my_add_history(char *line)
 {       
   int lookback, count, here;
-  char *new_entry, *filtered_line;
+  char *new_entry, *filtered_line, **lineptr, **list;
 
-  filtered_line =  pass_through_filter(TAG_HISTORY, line);
+  /* with bracketed-paste mode, line may actually be multiple lines. We'll treat each of those lines */
+  /* as a separate history item (I believe that is what bash does as well).                          */
+
+  list = split_with(line,"\n"); 
+  for (lineptr = list; *lineptr; lineptr++) {
+    filtered_line =  pass_through_filter(TAG_HISTORY, *lineptr);
  
   
-  switch (history_duplicate_avoidance_policy) { 
-  case KEEP_ALL_DOUBLES:
-    lookback = 0; break;
-  case ELIMINATE_SUCCESIVE_DOUBLES:
-    lookback = 1; break;
-  case ELIMINATE_ALL_DOUBLES:
-    lookback = history_length; break;
-  }
-
-  
-  new_entry = filtered_line;    
-  
-  lookback = min(history_length, lookback);      
-  for (count = 0, here = history_length - 1;
-       count < lookback ;
-       count++, here--) {
-    DPRINTF4(DEBUG_READLINE, "strcmping <%s> and <%s> (count = %d, here = %d)", line, history_get(history_base + here)->line ,count, here);
-    if (strncmp(new_entry, history_get(history_base + here) -> line, 10000) == 0) { /* history_get uses the logical offset history_base .. */
-       HIST_ENTRY *entry = remove_history (here);                                   /* .. but remove_history doesn't!                      */
-      DPRINTF2(DEBUG_READLINE, "removing duplicate entry #%d (%s)", here, entry->line);
-      free_foreign(entry->line);
-      free_foreign(entry);
+    switch (history_duplicate_avoidance_policy) { 
+    case KEEP_ALL_DOUBLES:
+      lookback = 0; break;
+    case ELIMINATE_SUCCESIVE_DOUBLES:
+      lookback = 1; break;
+    case ELIMINATE_ALL_DOUBLES:
+      lookback = history_length; break;
     }
-  }
-  add_history(new_entry);
-  free(new_entry);
-}
 
+  
+    new_entry = filtered_line;    
+  
+    lookback = min(history_length, lookback);      
+    for (count = 0, here = history_length - 1;
+         count < lookback ;
+         count++, here--) {
+      DPRINTF4(DEBUG_READLINE, "comparing <%s> and <%s> (count = %d, here = %d)", line
+             , history_get(history_base + here)->line ,count, here);
+      if (strncmp(new_entry, history_get(history_base + here) -> line, 10000) == 0) { /* history_get uses the logical offset history_base .. */
+        HIST_ENTRY *entry = remove_history (here);                                   /* .. but remove_history doesn't!                      */
+        DPRINTF2(DEBUG_READLINE, "removing duplicate entry #%d (%s)", here, entry->line);
+        free_foreign(entry->line);
+        free_foreign(entry);
+      }
+    }
+    add_history(new_entry);
+    free(new_entry);
+  }
+  free_splitlist(list);
+}
+  
 /* Homegrown redisplay function - erases current line and prints the
    new one.  Used for passwords (where we want to show **** instead of
    user input) and whenever HOMEGROWN_REDISPLAY is defined (for
@@ -368,7 +416,7 @@ my_homegrown_redisplay(int hide_passwords)
   int skip = max(1, min(width / 5, 10));        /* jumpscroll this many positions when cursor reaches edge of terminal */
   
   char *prompt_without_ignore_markers;
-  int colourless_promptlen = colourless_strlen(rl_prompt, &prompt_without_ignore_markers,0);
+  int colourless_promptlen = colourless_strlen(rl_prompt, &prompt_without_ignore_markers, 0, 0, NULL);
   int promptlen = strlen(prompt_without_ignore_markers);
   int invisible_chars_in_prompt = promptlen - colourless_promptlen;
   char *prompt_plus_line = add2strings(prompt_without_ignore_markers, rl_line_buffer);
@@ -486,7 +534,7 @@ my_homegrown_redisplay(int hide_passwords)
 
 
 void
-my_redisplay()
+my_redisplay(void)
 {
   int debug_force_homegrown_redisplay = 0;
 
@@ -616,12 +664,34 @@ munge_line_in_editor(int UNUSED(count), int UNUSED(key))
 static int
 direct_keypress(int UNUSED(count), int key)
 {
-  char *key_as_str = mysavestring("?");
-  /* put the key in the output queue    */
+#ifdef HAVE_RL_EXECUTING_KEYSEQ /* i.e. if readline version is >= 6.3 */
+  MAYBE_UNUSED(key);
+  DPRINTF1(DEBUG_READLINE,"direct keypress: %s", M(rl_executing_keyseq));
+  put_in_output_queue(rl_executing_keyseq);
+#else
+  char *key_as_str = mysavestring("?"); /* ? is just a placeholder */
   *key_as_str = key;
   DPRINTF1(DEBUG_READLINE,"direct keypress: %s", mangle_char_for_debug_log(key, TRUE));
   put_in_output_queue(key_as_str);
   free(key_as_str);
+#endif
+  return 0;
+}
+
+static int
+direct_prefix(int count, int key)
+{
+  char *key_as_str = mysavestring("?"); /* ? is just a placeholder */
+  
+  /* process the keypress used to invoke the function */
+  direct_keypress(count, key);
+  /* read an extra key to pass on */
+  key = rl_read_key();
+  *key_as_str = key;
+  DPRINTF1(DEBUG_READLINE,"read key: %s", mangle_char_for_debug_log(key, TRUE));
+  put_in_output_queue(key_as_str);
+  free(key_as_str);
+
   return 0;
 }
 
@@ -635,7 +705,7 @@ static char* entire_history_as_one_string(void) {
 
   for (entryp = the_list; *entryp; entryp++) {
     int length = strlen((*entryp)->line);
-    strncpy(stringp, (*entryp)->line, length); /* copy line, without closing NULL byte; */
+    strcpy(stringp, (*entryp)->line); 
     stringp +=length;
     *stringp++ = '\n';
   }
@@ -674,7 +744,7 @@ handle_hotkey2(int UNUSED(count), int hotkey, int without_history)
 #endif
 
     
-  DPRINTF3(DEBUG_READLINE, "hotkey press (without_history == %d): %x (%s)", without_history, hotkey, mangle_string_for_debug_log(executing_keyseq, MANGLE_LENGTH));
+  DPRINTF3(DEBUG_READLINE, "hotkey press (without_history == %d): %x (%s)", without_history, hotkey, M(executing_keyseq));
 
   prefix = mysavestring(rl_line_buffer);
   prefix[rl_point] = '\0';                                     /* chop off just before cursor */
@@ -786,10 +856,10 @@ colourise (const char *prompt)
 {
   char *prompt_copy, *trailing_space, *colour_end_with_space, *result, *p;
   prompt_copy = mysavestring(prompt);
-  if (strchr(prompt_copy, '\033') || strchr(prompt_copy, RL_PROMPT_START_IGNORE) ) {     /* prompt contains escape codes? */
-    DPRINTF1(DEBUG_READLINE, "colourise %s: left as-is", prompt);
-    return prompt_copy; /* if so, leave prompt alone  */
-  }
+  /* if (strchr(prompt_copy, '\033') || strchr(prompt_copy, RL_PROMPT_START_IGNORE) ) {     /\* prompt contains escape codes? *\/ */
+  /*   DPRINTF1(DEBUG_READLINE, "colourise %s: left as-is", prompt); */
+  /*   return prompt_copy; /\* if so, leave prompt alone  *\/ */
+  /* } */
   for (p = prompt_copy + strlen(prompt_copy); p > prompt_copy && *(p-1) == ' '; p--)
     ; /* skip back over trailing space */
   trailing_space = mysavestring(p); /* p now points at first trailing space, or else the final NULL */
@@ -838,10 +908,6 @@ prompt_is_single_line(void)
 #ifndef SPY_ON_READLINE
 #  define _rl_horizontal_scroll_mode FALSE
 #  define rl_variable_value(s) "off"
-#else
-#  ifndef HAVE_RL_VARIABLE_VALUE
-#    define rl_variable_value(s) "off"
-#  endif
 #endif
   
 #ifdef HOMEGROWN_REDISPLAY
@@ -890,8 +956,10 @@ char *process_new_output(const char* buffer, struct rl_state* UNUSED(state)) {
 }
 
 
-int cook_prompt_if_necessary () {
-  char *pre_cooked, *rubbish_from_alternate_screen,  *filtered, *uncoloured, *cooked, *p, *non_rubbish = NULL;
+
+
+int cook_prompt_if_necessary (void) {
+  char *pre_cooked, *slightly_cooked, *rubbish_from_alternate_screen,  *filtered, *uncoloured, *cooked, *p, *non_rubbish = NULL;
   static char **term_ctrl_seqs[] 
     = {&term_rmcup, &term_rmkx, NULL}; /* (NULL-terminated) list of (pointers to) term control sequences that may be
                                        used by clients to return from an 'alternate screen'. If we spot one of those,
@@ -926,22 +994,24 @@ int cook_prompt_if_necessary () {
     rubbish_from_alternate_screen = mysavestring("");
   }
   
-
-  unbackspace(pre_cooked); /* programs that display a running counter would otherwise make rlwrap keep prompts
+  slightly_cooked = ansi_colour_aware ? protect_or_cleanup(pre_cooked, FALSE) : mysavestring(pre_cooked);
+  free(pre_cooked);
+  
+  unbackspace(slightly_cooked); /* programs that display a running counter would otherwise make rlwrap keep prompts
                               like " 1%\r 2%\r 3%\ ......" */
 
   if ( /* raw prompt doesn't match '--only-cook' regexp */
-      (prompt_regexp && ! match_regexp(pre_cooked, prompt_regexp, FALSE)) ||
+      (prompt_regexp && ! match_regexp(slightly_cooked, prompt_regexp, FALSE)) ||
        /* now filter it, but filter may "refuse" the prompt */
-      (strcmp((filtered =  pass_through_filter(TAG_PROMPT, pre_cooked)), "_THIS_CANNOT_BE_A_PROMPT_")== 0)) { 
+      (strcmp((filtered =  pass_through_filter(TAG_PROMPT, slightly_cooked)), "_THIS_CANNOT_BE_A_PROMPT_")== 0)) { 
     /* don't cook, eat raw (and eat nothing if patient) */       
-    saved_rl_state.cooked_prompt =  (impatient_prompt ? mysavestring(pre_cooked) : mysavestring("")); 
+    saved_rl_state.cooked_prompt =  (impatient_prompt ? mysavestring(slightly_cooked) : mysavestring("")); 
     /* NB: if impatient, the rubbish_from_alternate_screen has been output already, no need to send it again */  
-    free(pre_cooked);
+    free(slightly_cooked);
     free(filtered); /* free(NULL) is never an error */
     return FALSE;
-  }     
-  free(pre_cooked);
+  }   
+  free(slightly_cooked);
   if(substitute_prompt) {
     uncoloured = mysavestring(substitute_prompt); 
     free(filtered);
@@ -955,9 +1025,10 @@ int cook_prompt_if_necessary () {
     cooked = uncoloured;
   }
   if (! impatient_prompt)  /* in this case our rubbish hasn't been output yet. Output it now, but don't store
-                              it in the prompt, as this may be re-printed e.g. after resuming a suspended rlwrap */
-                              
+                              it in the prompt, as this may be re-printed e.g. after resuming a suspended rlwrap */                            
     write_patiently(STDOUT_FILENO,rubbish_from_alternate_screen, strlen(rubbish_from_alternate_screen), "to stdout");
+
+  free(rubbish_from_alternate_screen);
   saved_rl_state.cooked_prompt = cooked;
   return TRUE;
 }       
@@ -997,15 +1068,54 @@ static Keymap getmap(const char *name) {
   return km;
 }
 
-static void bindkey(int key, rl_command_func_t *function, const char *maplist) {
+static void bindkey(int key, rl_command_func_t *function, const char *function_name, const char *maplist) {
   char *mapnames[] = {"emacs-standard","emacs-ctlx","emacs-meta","vi-insert","vi-move","vi-command",NULL};
   char **mapname;
   for (mapname = mapnames; *mapname; mapname++) 
     if(strstr(maplist, *mapname)) {
       Keymap kmap = getmap(*mapname);
-      DPRINTF4(DEBUG_READLINE,"Binding key %d (%s) in keymap '%s' to <0x%lx>", key, mangle_char_for_debug_log(key,TRUE), *mapname, (long) function);
+      DPRINTF4(DEBUG_READLINE,"Binding key %d (%s) in keymap '%s' to %s()", key, mangle_char_for_debug_log(key,TRUE), *mapname, function_name);
       if (rl_bind_key_in_map(key, function, kmap))
         myerror(FATAL|NOERRNO, "Could not bind key %d (%s) in keymap '%s'", key, mangle_char_for_debug_log(key,TRUE), *mapname);
     }
 }       
 
+
+#ifdef UNIT_TEST
+
+void my_line_handler(char *line) { /* This function (callback) gets called by readline
+                                   whenever rl_callback_read_char sees an ENTER */ 
+  printf("\nYou typed: '%s'\n", line);
+  if (!line)
+    exit(0);
+}
+      
+
+TESTFUNC(simple_callback, argc, argv, stage) {
+  char byte_read;
+  ONLY_AT_STAGE(TEST_AFTER_OPTION_PARSING);
+  init_readline("Weh! ");
+  set_echo(FALSE);
+  rl_callback_handler_install("Type anything: ", &my_line_handler);
+  while (TRUE) {
+    if (!read(STDIN_FILENO, &byte_read, 1))
+      exit(0);
+    DPRINTF1(DEBUG_READLINE, "Byte read: %s", mangle_char_for_debug_log(byte_read, TRUE));
+    rl_stuff_char(byte_read);
+    rl_callback_read_char();
+  }
+ }    
+
+TESTFUNC(standard_readline, argc, argv, stage) {
+  ONLY_AT_STAGE(TEST_AFTER_OPTION_PARSING);
+  while(TRUE) {
+     char *line_read = readline ("go ahead: ");
+     printf("\nYou typed: '%s'\n", line_read);
+     if (!line_read || !*line_read)
+       break;
+     free(line_read);
+  }
+  exit(0);
+}       
+
+#endif /* UNIT_TEST */
